#!/usr/bin/python3

from random import shuffle

from player import Player
import ai


def spies_for_players(player_count):
    return {
        5: 2,
        6: 2,
        7: 3,
        8: 3,
        9: 3,
        10: 4
    }[player_count]


def mission_size(player_count, mission_num):
    return {
        5: [2, 3, 2, 3, 3],
        6: [2, 3, 4, 3, 4],
        7: [2, 3, 3, 4, 4],
        8: [3, 4, 4, 5, 5],
        9: [3, 4, 4, 5, 5],
        10: [3, 4, 4, 5, 5]
    }[player_count][mission_num - 1]


def required_failures(player_count, mission_num):
    if mission_num == 4 and player_count >= 7:
        return 2
    else:
        return 1


class Shared(object):

    def __init__(self, player_count, players):
        self.player_count = player_count
        self.players = players
        self.mission = 0
        self.captain = 0
        self.team_failures = 0
        self.wins = 0


class AdvanceMission(object):

    def __init__(self, shared):
        self.shared = shared

    def do(self):
        self.shared.mission += 1
        team_size = mission_size(self.shared.player_count, self.shared.mission)
        print("Mission {} needs {} members".format(self.shared.mission, team_size))
        return AssembleTeam(self.shared, team_size)


class AssembleTeam(object):

    def __init__(self, shared, team_size):
        self.shared = shared
        self.team_size = team_size

    def do(self):
        print("Captain is {}".format(self.shared.captain))
        team = self.shared.players[self.shared.captain].propose_team(self.team_size)
        print("He proposed {}".format(team))

        votes = [p.approve_team(team) for p in self.shared.players]
        approvals = sum(votes)
        approved = approvals > self.team_size / 2

        print("Team {}: {} for, {} against".format(
            {True: "approved", False: "rejected"}[approved],
            approvals, self.shared.player_count - approvals
        ))

        for p in self.shared.players:
            p.observe_team_vote(approved, votes)

        self.shared.captain = (self.shared.captain + 1) % self.shared.player_count

        if approved:
            self.shared.team_failures = 0
            return PerformMission(self.shared, team)
        else:
            self.shared.team_failures += 1
            return AssembleTeam(self.shared, self.team_size)


class PerformMission(object):

    def __init__(self, shared, team):
        self.shared = shared
        self.team = team


def main(player_count):
    spy_count = spies_for_players(player_count)
    resistance_count = player_count - spy_count
    print("Will have {} spies and {} resistance".format(spy_count, resistance_count))
    role_deck = ["RES"] * resistance_count + ["SPY"] * spy_count
    shuffle(role_deck)
    print(role_deck)
    players = [Player(player_count, i, r) if r == "SPY"
               else ai.Player(player_count, i, r)
               for i, r in enumerate(role_deck)]

    spy_nums = [i for i, r in enumerate(role_deck) if r == "SPY"]

    for i, r in enumerate(role_deck):
        if r == "SPY":
            players[i].reveal_spies(spy_nums)

    shared = Shared(player_count, players)

    losses = 0

    def perform_mission():
        nonlocal losses

        failures = 0
        for i in state.team:
            if role_deck[i] == "SPY":
                if not state.shared.players[i].perform_mission():
                    failures = failures + 1

        success = failures < required_failures(state.shared.player_count, state.shared.mission)

        for p in state.shared.players:
            p.observe_mission(success, failures)

        if success:
            print("Mission succeeded")
            state.shared.wins += 1
        else:
            print("Mission had {} failures".format(failures))
            losses += 1

        return AdvanceMission(shared)

    state = AdvanceMission(shared)

    while True:
        if isinstance(state, AdvanceMission):
            state = state.do()
        elif isinstance(state, AssembleTeam):
            state = state.do()
        elif isinstance(state, PerformMission):
            state = perform_mission()

        if shared.team_failures == 5:
            print("The spies won!")
            for p in shared.players:
                p.observe_game(False)
            return
        elif shared.wins >= 3:
            print("The Resistance won!")
            for p in shared.players:
                p.observe_game(True)
            return
        elif losses >= 3:
            print("The Spies won!")
            for p in shared.players:
                p.observe_game(False)
            return


if __name__ == '__main__':
    main(10)
